@@ -1,38 +1,40 @@
 """A package for encoding and decoding weather dataset."""
 
+import asyncio
+import copy
 import io
 import os
 import pickle
-import copy
+import random
+import threading
+import time
+from collections import deque
+from functools import partial
+from time import perf_counter
 from typing import (
+    Any,
     BinaryIO,
+    Deque,
     Dict,
     Generator,
     List,
     Optional,
-    Deque,
     Tuple,
     TypeVar,
     Union,
-    Any,
 )
-from time import perf_counter
-import numpy as np
-import random
-import yaml
-import threading
-import time
-import asyncio
+
 import aiofiles
-from functools import partial
-from collections import deque
+import numpy as np
+import yaml
+
 from seqrecord.utils import (
-    LRUCache,
-    FileManager,
-    WriterBuffer,
-    TimeTracker,
-    PRODUCER_SLEEP_INTERVAL,
     CONSUMER_SLEEP_INTERVAL,
+    PRODUCER_SLEEP_INTERVAL,
+    FileManager,
+    LRUCache,
+    TimeTracker,
+    WriterBuffer,
 )
 
 MAX_RECORDFILE_SIZE = 1e9  # 1e8, 100 mb, maximum size of a single record file
@@ -139,6 +141,7 @@ class WSeqRecord:
             self.write_buffer.write(data.tobytes())
             num_bytes_in_frame += data.nbytes
 
+        self.metadata[self.frame_idx]["nbytes"] = num_bytes_in_frame
         self.num_bytes += num_bytes_in_frame
         self.frame_idx += 1
 
@@ -221,11 +224,12 @@ class WSeqRecord:
 
             if len(self.idx_range_of_files[-1]) == 1:
                 # there is content left in the write_buffer
-                self.idx_range_of_files[-1].append(self.frame_idx)
+                self.idx_range_of_files[-1].append(frame_idx)
                 yield (file_idx, write_buffer.getvalue())
         finally:
             write_buffer.close()
             self.num_files = len(self.idx_range_of_files)
+            self.num_frames = frame_idx
 
     def put_frame(self, frame_generator: callable, prefetch_buffer_size: int = 5):
         # should be only adding frames here
@@ -275,7 +279,6 @@ class WSeqRecord:
         frame = {}
         frame_offset = metadata_frame["bytes_offset"]
         for feature in features:
-            start = perf_counter()
             frame[feature] = np.memmap(
                 file_desc,
                 dtype=metadata_frame[feature]["dtype"],
@@ -283,8 +286,6 @@ class WSeqRecord:
                 offset=frame_offset + metadata_frame[feature]["bytes_offset"],
                 shape=metadata_frame[feature]["shape"],
             )
-            end = perf_counter()
-            self.timer.add(end - start, frame[feature].nbytes)
         return frame
 
     def iterate_frames(
@@ -346,13 +347,14 @@ class WSeqRecord:
         assert all(
             len(endpoints) == 2 for endpoints in self.idx_range_of_files
         ), "Record file endpoints are not complete!"
-
+        num_frames = len(self.metadata)  # or self.num_frames
         for fileidx4input in range(self.num_files):
             filedesc4input = file_manager.open_file(file_idx=fileidx4input)
             endpoints = self.idx_range_of_files[fileidx4input]
             # no target frame to predict for the last frame
             for frameidx4input in range(
-                endpoints[0], min(endpoints[1], self.num_frames - 1)
+                endpoints[0],
+                min(endpoints[1], num_frames - 1),  # self.num_frames
             ):
                 input_frame = self.read_frame(
                     filedesc4input, self.metadata[frameidx4input], input_features
@@ -360,7 +362,7 @@ class WSeqRecord:
                 # get the target frame for prediction, both start, stop inclusive
                 lookahead_steps = min(
                     random.randint(1, max_pred_steps),
-                    self.num_frames - 1 - frameidx4input,
+                    num_frames - 1 - frameidx4input,
                 )
                 frameidx4target = frameidx4input + lookahead_steps
                 target_frame = frame_cache.get(frameidx4target)
@@ -419,20 +421,21 @@ class WSeqRecord:
             len(endpoints) == 2 for endpoints in self.idx_range_of_files
         ), "Record file endpoints are not complete!"
 
+        num_frames = len(self.metadata)
         # read two frames using asyn io
         # file cache should only be used for future frames, since base file desc is used continuously
         try:
             for fileidx4input in range(self.num_files):
                 filedesc4input = None
                 endpoints = self.idx_range_of_files[fileidx4input]
+
                 # no target frame to predict for the last frame
                 for frameidx4input in range(
-                    endpoints[0], min(endpoints[1], self.num_frames - 1)
+                    endpoints[0], min(endpoints[1], num_frames - 1)
                 ):
-
                     lookahead_steps = min(
                         random.randint(1, max_pred_steps),
-                        self.num_frames - 1 - frameidx4input,
+                        num_frames - 1 - frameidx4input,
                     )
                     frameidx4target = frameidx4input + lookahead_steps
                     fileidx4target = self.metadata[frameidx4target]["file_idx"]
@@ -494,7 +497,9 @@ class WSeqRecord:
                         "target_features": target_features,
                     }
                 # close file descriptor for
-                loop.run_until_complete(close_aiofile(filedesc4input))
+
+                if filedesc4input is not None:
+                    loop.run_until_complete(close_aiofile(filedesc4input))
         finally:
             # close open files
             loop.run_until_complete(close_files_in_cache(file_desc_cache))
@@ -640,20 +645,6 @@ async def close_aiofile(file_desc: io.BufferedReader) -> None:
     return
 
 
-async def read_chunk(
-    file_desc: io.BufferedReader, frame_offset: int, metadata_feature: dict
-) -> np.ndarray:
-    offset = frame_offset + metadata_feature["bytes_offset"]
-    # do I need to make sure there is no interruption between seek and read?
-    await file_desc.seek(offset)
-    data = await file_desc.read(metadata_feature["nbytes"])
-    array1d = np.frombuffer(
-        data,
-        dtype=metadata_feature["dtype"],
-    )
-    return array1d.reshape(metadata_feature["shape"])
-
-
 # notes: have to use file manager (instead of an lru function since we need to close files)
 #        how to verify we are doing async?
 # other approaches to be compared with:
@@ -673,6 +664,8 @@ async def async_read_frame(
     Returns:
         np.ndarray: _description_
     """
+    await file_desc.seek(metadata_frame["bytes_offset"])
+    data_bytes = await file_desc.read(metadata_frame["nbytes"])
     frame = {}
     # read the whole chunk or we read each file separately
     for feature in features:
@@ -681,11 +674,14 @@ async def async_read_frame(
         #     bytes,
         #     dtype=metadata[feature]["dtype"],
         # )
-        # frame[feature] = array1d.reshape(metadata[feature]["shape"])
+        # frame[feature] = array1d
         # `await` halts `async_read_frame` and gives control back
-        frame[feature] = await read_chunk(
-            file_desc, metadata_frame["bytes_offset"], metadata_frame[feature]
-        )
+        start = metadata_frame[feature]["bytes_offset"]
+        end = start + metadata_frame[feature]["nbytes"]
+        frame[feature] = np.frombuffer(
+            data_bytes[start:end],
+            dtype=metadata_frame[feature]["dtype"],
+        ).reshape(metadata_frame[feature]["shape"])
 
     frame_array = np.vstack([frame[feature] for feature in features])
     return frame_array
